/*
 * Copyright (C) 2011 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package com.google.android.marvin.talkback;

import android.content.Context;
import android.os.Build;
import android.os.Bundle;
import android.os.Message;
import android.support.v4.view.accessibility.AccessibilityEventCompat;
import android.support.v4.view.accessibility.AccessibilityNodeInfoCompat;
import android.support.v4.view.accessibility.AccessibilityRecordCompat;
import android.text.TextUtils;
import android.view.accessibility.AccessibilityEvent;

import com.googlecode.eyesfree.compat.view.accessibility.AccessibilityEventCompatUtils;
import com.googlecode.eyesfree.utils.AccessibilityEventListener;
import com.googlecode.eyesfree.utils.AccessibilityEventUtils;
import com.googlecode.eyesfree.utils.AccessibilityNodeInfoUtils;
import com.googlecode.eyesfree.utils.WeakReferenceHandler;

import java.util.HashMap;

/**
 * Manages scroll position feedback. If a VIEW_SCROLLED event passes through
 * this processor and no further events are received for a specified duration, a
 * "scroll position" message is spoken.
 *
 * @author alanv@google.com (Alan Viverette)
 */
class ProcessorScrollPosition implements AccessibilityEventListener {
    /** Default pitch adjustment for text event feedback. */
    private static final float DEFAULT_PITCH = 1.2f;

    /** Default rate adjustment for text event feedback. */
    private static final float DEFAULT_RATE = 1.0f;

    private final HashMap<AccessibilityNodeInfoCompat, Integer>
            mCachedFromValues = new HashMap<AccessibilityNodeInfoCompat, Integer>();
    private final Bundle mSpeechParams = new Bundle();

    private final Context mContext;
    private final SpeechController mSpeechController;
    private final FullScreenReadController mFullScreenReadController;

    private AccessibilityNodeInfoCompat mRecentlyExplored;

    public ProcessorScrollPosition(TalkBackService context) {
        mContext = context;
        mSpeechController = context.getSpeechController();

        if (Build.VERSION.SDK_INT >= FullScreenReadController.MIN_API_LEVEL) {
            mFullScreenReadController = context.getFullScreenReadController();
        } else {
            mFullScreenReadController = null;
        }

        mSpeechParams.putFloat(SpeechController.SpeechParam.PITCH, DEFAULT_PITCH);
        mSpeechParams.putFloat(SpeechController.SpeechParam.RATE, DEFAULT_RATE);
    }

    @Override
    public void onAccessibilityEvent(AccessibilityEvent event) {
        updateRecentlyExplored(event);

        if (shouldIgnoreEvent(event)) {
            return;
        }

        switch (event.getEventType()) {
            case AccessibilityEvent.TYPE_WINDOW_STATE_CHANGED:
                // Window state changes clear the cache.
                mCachedFromValues.clear();
                break;
        }
    }

    private void updateRecentlyExplored(AccessibilityEvent event) {
        // TODO(alanv): We track this in several places. Need to refactor.

        // We only need to track recently explored views on ICS.
        if ((Build.VERSION.SDK_INT < Build.VERSION_CODES.ICE_CREAM_SANDWICH)
                || (Build.VERSION.SDK_INT >= Build.VERSION_CODES.JELLY_BEAN)) {
            return;
        }

        // Exploration is the result of HOVER_ENTER events.
        if (event.getEventType() != AccessibilityEventCompat.TYPE_VIEW_HOVER_ENTER) {
            return;
        }

        final AccessibilityRecordCompat record = new AccessibilityRecordCompat(event);
        final AccessibilityNodeInfoCompat node = record.getSource();

        if (mRecentlyExplored != null) {
            if (mRecentlyExplored.equals(node)) {
                return;
            }

            mRecentlyExplored.recycle();
            mRecentlyExplored = null;
        }

        if (node != null) {
            mRecentlyExplored = AccessibilityNodeInfoCompat.obtain(node);
        }
    }

    private boolean shouldIgnoreEvent(AccessibilityEvent event) {
        switch (event.getEventType()) {
            case AccessibilityEventCompat.TYPE_WINDOW_CONTENT_CHANGED:
            case AccessibilityEventCompat.TYPE_VIEW_ACCESSIBILITY_FOCUSED:
            case AccessibilityEventCompat.TYPE_VIEW_ACCESSIBILITY_FOCUS_CLEARED:
                return true;
            case AccessibilityEventCompat.TYPE_VIEW_SCROLLED:
                return shouldIgnoreScrollEvent(event);
            default:
                return false;
        }
    }

    private boolean shouldIgnoreScrollEvent(AccessibilityEvent event) {
        // Don't speak during full-screen read.
        if ((mFullScreenReadController != null) && mFullScreenReadController.isActive()) {
            return true;
        }

        final AccessibilityRecordCompat record = new AccessibilityRecordCompat(event);
        final AccessibilityNodeInfoCompat node = record.getSource();

        if (node == null) {
            if (Build.VERSION.SDK_INT >= 14) {
                // Event is coming from another window.
                return true;
            } else {
                // We don't have enough information to ignore this event.
                return false;
            }
        }

        final int fromIndex = event.getFromIndex() + 1;
        final Integer cachedFromIndex = mCachedFromValues.get(node);

        if ((cachedFromIndex != null) && (cachedFromIndex == fromIndex)) {
            // The from index hasn't changed, which means the event is coming
            // from a re-layout or resize and should not be spoken.
            node.recycle();
            return true;
        }

        // The behavior of put() for an existing key is unspecified, so we can't
        // recycle the old or new key nodes.
        mCachedFromValues.put(node, fromIndex);
        return false;
    }
}
